@@ -262,6 +262,10 @@ struct ServerOptions {
     // Default: ""
     std::string server_info_name;
 
+    // Server will run in this tagged bthread worker group
+    // Default: BTHREAD_TAG_DEFAULT
+    bthread_tag_t bthread_tag;
+
 private:
     // SSLOptions is large and not often used, allocate it on heap to
     // prevent ServerOptions from being bloated in most cases.
@@ -339,7 +343,7 @@ struct PortRange {
 };
 
 // Server dispatches requests from clients to registered services and
-// and sends responses back to clients.
+// sends responses back to clients.
 class Server {
 public:
     enum Status {
@@ -660,7 +664,7 @@ friend class Controller;
     void FreeSSLContexts();
 
     static int SSLSwitchCTXByHostname(struct ssl_st* ssl,
-                                      int* al, Server* server);
+                                      int* al, void* se);
 
     static bool AddCertMapping(CertMaps& bg, const SSLContext& ssl_ctx);
     static bool RemoveCertMapping(CertMaps& bg, const SSLContext& ssl_ctx);
