@@ -689,7 +689,7 @@ int StreamWrite(StreamId stream_id, const butil::IOBuf &message,
         return EINVAL;
     }
     Stream* s = (Stream*)ptr->conn();
-    const int rc = s->AppendIfNotFull(message);
+    const int rc = s->AppendIfNotFull(message, options);
     if (rc == 0) {
         return 0;
     }
@@ -761,7 +761,7 @@ int StreamAccept(StreamId* response_stream, Controller &cntl,
                  const StreamOptions* options) {
 
     if (cntl._response_stream != INVALID_STREAM_ID) {
-        LOG(ERROR) << "Can't create reponse stream more than once";
+        LOG(ERROR) << "Can't create response stream more than once";
         return -1;
     }
     if (response_stream == NULL) {
