@@ -38,7 +38,13 @@ DEFINE_int32(bthread_min_concurrency, 0,
             " The laziness is disabled when this value is non-positive,"
             " and workers will be created eagerly according to -bthread_concurrency and bthread_setconcurrency(). ");
 
+DEFINE_int32(bthread_current_tag, BTHREAD_TAG_DEFAULT, "Set bthread concurrency for this tag");
+
+DEFINE_int32(bthread_concurrency_by_tag, 0,
+             "Number of pthread workers of FLAGS_bthread_current_tag");
+
 static bool never_set_bthread_concurrency = true;
+static bool never_set_bthread_concurrency_by_tag = true;
 
 static bool validate_bthread_concurrency(const char*, int32_t val) {
     // bthread_setconcurrency sets the flag on success path which should
@@ -55,6 +61,17 @@ const int ALLOW_UNUSED register_FLAGS_bthread_min_concurrency =
     ::GFLAGS_NS::RegisterFlagValidator(&FLAGS_bthread_min_concurrency,
                                     validate_bthread_min_concurrency);
 
+static bool validate_bthread_current_tag(const char*, int32_t val);
+
+const int ALLOW_UNUSED register_FLAGS_bthread_current_tag =
+    ::GFLAGS_NS::RegisterFlagValidator(&FLAGS_bthread_current_tag, validate_bthread_current_tag);
+
+static bool validate_bthread_concurrency_by_tag(const char*, int32_t val);
+
+const int ALLOW_UNUSED register_FLAGS_bthread_concurrency_by_tag =
+    ::GFLAGS_NS::RegisterFlagValidator(&FLAGS_bthread_concurrency_by_tag,
+                                       validate_bthread_concurrency_by_tag);
+
 BAIDU_CASSERT(sizeof(TaskControl*) == sizeof(butil::atomic<TaskControl*>), atomic_size_match);
 
 pthread_mutex_t g_task_control_mutex = PTHREAD_MUTEX_INITIALIZER;
@@ -65,6 +82,7 @@ TaskControl* g_task_control = NULL;
 
 extern BAIDU_THREAD_LOCAL TaskGroup* tls_task_group;
 extern void (*g_worker_startfn)();
+extern void (*g_tagged_worker_startfn)(bthread_tag_t);
 
 inline TaskControl* get_task_control() {
     return g_task_control;
@@ -97,6 +115,15 @@ inline TaskControl* get_or_new_task_control() {
     return c;
 }
 
+static int add_workers_for_each_tag(int num) {
+    int added = 0;
+    auto c = get_task_control();
+    for (auto i = 0; i < num; ++i) {
+        added += c->add_workers(1, i % FLAGS_task_group_ntags);
+    }
+    return added;
+}
+
 static bool validate_bthread_min_concurrency(const char*, int32_t val) {
     if (val <= 0) {
         return true;
@@ -111,38 +138,68 @@ static bool validate_bthread_min_concurrency(const char*, int32_t val) {
     BAIDU_SCOPED_LOCK(g_task_control_mutex);
     int concurrency = c->concurrency();
     if (val > concurrency) {
-        int added = c->add_workers(val - concurrency);
+        int added = bthread::add_workers_for_each_tag(val - concurrency);
         return added == (val - concurrency);
     } else {
         return true;
     }
 }
 
+static bool validate_bthread_current_tag(const char*, int32_t val) {
+    if (val < BTHREAD_TAG_DEFAULT || val >= FLAGS_task_group_ntags) {
+        return false;
+    }
+    BAIDU_SCOPED_LOCK(bthread::g_task_control_mutex);
+    auto c = bthread::get_task_control();
+    if (c == NULL) {
+        FLAGS_bthread_concurrency_by_tag = 0;
+        return true;
+    }
+    FLAGS_bthread_concurrency_by_tag = c->concurrency(val);
+    return true;
+}
+
+static bool validate_bthread_concurrency_by_tag(const char*, int32_t val) {
+    return bthread_setconcurrency_by_tag(val, FLAGS_bthread_current_tag) == 0;
+}
+
 __thread TaskGroup* tls_task_group_nosignal = NULL;
 
 BUTIL_FORCE_INLINE int
 start_from_non_worker(bthread_t* __restrict tid,
                       const bthread_attr_t* __restrict attr,
-                      void * (*fn)(void*),
+                      void* (*fn)(void*),
                       void* __restrict arg) {
     TaskControl* c = get_or_new_task_control();
     if (NULL == c) {
         return ENOMEM;
     }
+    auto tag = BTHREAD_TAG_DEFAULT;
+    if (attr != NULL && attr->tag != BTHREAD_TAG_INVALID) {
+        tag = attr->tag;
+    }
     if (attr != NULL && (attr->flags & BTHREAD_NOSIGNAL)) {
         // Remember the TaskGroup to insert NOSIGNAL tasks for 2 reasons:
         // 1. NOSIGNAL is often for creating many bthreads in batch,
         //    inserting into the same TaskGroup maximizes the batch.
         // 2. bthread_flush() needs to know which TaskGroup to flush.
-        TaskGroup* g = tls_task_group_nosignal;
+        auto g = tls_task_group_nosignal;
         if (NULL == g) {
-            g = c->choose_one_group();
+            g = c->choose_one_group(tag);
             tls_task_group_nosignal = g;
         }
         return g->start_background<true>(tid, attr, fn, arg);
     }
-    return c->choose_one_group()->start_background<true>(
-        tid, attr, fn, arg);
+    return c->choose_one_group(tag)->start_background<true>(tid, attr, fn, arg);
+}
+
+// Meet one of the three conditions, can run in thread local
+// attr is nullptr
+// tag equal to thread local
+// tag equal to BTHREAD_TAG_INVALID
+BUTIL_FORCE_INLINE bool can_run_thread_local(const bthread_attr_t* __restrict attr) {
+    return attr == nullptr || attr->tag == bthread::tls_task_group->tag() ||
+           attr->tag == BTHREAD_TAG_INVALID;
 }
 
 struct TidTraits {
@@ -175,8 +232,10 @@ int bthread_start_urgent(bthread_t* __restrict tid,
                          void* __restrict arg) {
     bthread::TaskGroup* g = bthread::tls_task_group;
     if (g) {
-        // start from worker
-        return bthread::TaskGroup::start_foreground(&g, tid, attr, fn, arg);
+        // if attribute is null use thread local task group
+        if (bthread::can_run_thread_local(attr)) {
+            return bthread::TaskGroup::start_foreground(&g, tid, attr, fn, arg);
+        }
     }
     return bthread::start_from_non_worker(tid, attr, fn, arg);
 }
@@ -187,8 +246,10 @@ int bthread_start_background(bthread_t* __restrict tid,
                              void* __restrict arg) {
     bthread::TaskGroup* g = bthread::tls_task_group;
     if (g) {
-        // start from worker
-        return g->start_background<false>(tid, attr, fn, arg);
+        // if attribute is null use thread local task group
+        if (bthread::can_run_thread_local(attr)) {
+            return g->start_background<false>(tid, attr, fn, arg);
+        }
     }
     return bthread::start_from_non_worker(tid, attr, fn, arg);
 }
@@ -306,13 +367,47 @@ int bthread_setconcurrency(int num) {
     }
     if (num > bthread::FLAGS_bthread_concurrency) {
         // Create more workers if needed.
-        bthread::FLAGS_bthread_concurrency +=
-            c->add_workers(num - bthread::FLAGS_bthread_concurrency);
-        return 0;
+        auto added = bthread::add_workers_for_each_tag(num - bthread::FLAGS_bthread_concurrency);
+        bthread::FLAGS_bthread_concurrency += added;
     }
     return (num == bthread::FLAGS_bthread_concurrency ? 0 : EPERM);
 }
 
+int bthread_getconcurrency_by_tag(bthread_tag_t tag) {
+    BAIDU_SCOPED_LOCK(bthread::g_task_control_mutex);
+    auto c = bthread::get_task_control();
+    if (c == NULL) {
+        return EPERM;
+    }
+    return c->concurrency(tag);
+}
+
+int bthread_setconcurrency_by_tag(int num, bthread_tag_t tag) {
+    if (bthread::never_set_bthread_concurrency_by_tag) {
+        bthread::never_set_bthread_concurrency_by_tag = false;
+        return 0;
+    }
+    BAIDU_SCOPED_LOCK(bthread::g_task_control_mutex);
+    auto c = bthread::get_task_control();
+    if (c == NULL) {
+        return EPERM;
+    }
+    auto ngroup = c->concurrency();
+    auto tag_ngroup = c->concurrency(tag);
+    auto add = num - tag_ngroup;
+    if (ngroup + add > bthread::FLAGS_bthread_concurrency) {
+        LOG(ERROR) << "Fail to set concurrency by tag " << tag
+                   << ", Whole concurrency larger than bthread_concurrency";
+        return EPERM;
+    }
+    auto added = 0;
+    if (add > 0) {
+        added = c->add_workers(add, tag);
+        return (add == added ? 0 : EPERM);
+    }
+    return (num == tag_ngroup ? 0 : EPERM);
+}
+
 int bthread_about_to_quit() {
     bthread::TaskGroup* g = bthread::tls_task_group;
     if (g != NULL) {
@@ -384,6 +479,14 @@ int bthread_set_worker_startfn(void (*start_fn)()) {
     return 0;
 }
 
+int bthread_set_tagged_worker_startfn(void (*start_fn)(bthread_tag_t)) {
+    if (start_fn == NULL) {
+        return EINVAL;
+    }
+    bthread::g_tagged_worker_startfn = start_fn;
+    return 0;
+}
+
 void bthread_stop_world() {
     bthread::TaskControl* c = bthread::get_task_control();
     if (c != NULL) {
@@ -433,5 +536,10 @@ int bthread_list_join(bthread_list_t* list) {
     static_cast<bthread::TidList*>(list->impl)->apply(bthread::TidJoiner());
     return 0;
 }
-    
+
+bthread_tag_t bthread_self_tag(void) {
+    return bthread::tls_task_group != nullptr ? bthread::tls_task_group->tag()
+                                              : BTHREAD_TAG_DEFAULT;
+}
+
 }  // extern "C"
