@@ -39,6 +39,7 @@ DEFINE_int32(task_group_runqueue_capacity, 4096,
              "capacity of runqueue in each TaskGroup");
 DEFINE_int32(task_group_yield_before_idle, 0,
              "TaskGroup yields so many times before idle");
+DEFINE_int32(task_group_ntags, 1, "TaskGroup will be grouped by number ntags");
 
 namespace bthread {
 
@@ -48,6 +49,7 @@ DECLARE_int32(bthread_min_concurrency);
 extern pthread_mutex_t g_task_control_mutex;
 extern BAIDU_THREAD_LOCAL TaskGroup* tls_task_group;
 void (*g_worker_startfn)() = NULL;
+void (*g_tagged_worker_startfn)(bthread_tag_t) = NULL;
 
 // May be called in other modules to run startfn in non-worker pthreads.
 void run_worker_startfn() {
@@ -56,28 +58,44 @@ void run_worker_startfn() {
     }
 }
 
+void run_tagged_worker_startfn(bthread_tag_t tag) {
+    if (g_tagged_worker_startfn) {
+        g_tagged_worker_startfn(tag);
+    }
+}
+
+struct WorkerThreadArgs {
+    WorkerThreadArgs(TaskControl* _c, bthread_tag_t _t) : c(_c), tag(_t) {}
+    TaskControl* c;
+    bthread_tag_t tag;
+};
+
 void* TaskControl::worker_thread(void* arg) {
-    run_worker_startfn();    
+    run_worker_startfn();
 #ifdef BAIDU_INTERNAL
     logging::ComlogInitializer comlog_initializer;
 #endif
-    
-    TaskControl* c = static_cast<TaskControl*>(arg);
-    TaskGroup* g = c->create_group();
+
+    auto dummy = static_cast<WorkerThreadArgs*>(arg);
+    auto c = dummy->c;
+    auto tag = dummy->tag;
+    delete dummy;
+    run_tagged_worker_startfn(tag);
+
+    TaskGroup* g = c->create_group(tag);
     TaskStatistics stat;
     if (NULL == g) {
         LOG(ERROR) << "Fail to create TaskGroup in pthread=" << pthread_self();
         return NULL;
     }
     std::string worker_thread_name = butil::string_printf(
-        "brpc_worker:%d",
-        c->_next_worker_id.fetch_add(1, butil::memory_order_relaxed));
+        "brpc_wkr:%d-%d", g->tag(), c->_next_worker_id.fetch_add(1, butil::memory_order_relaxed));
     butil::PlatformThread::SetName(worker_thread_name.c_str());
-    BT_VLOG << "Created worker=" << pthread_self()
-            << " bthread=" << g->main_tid();
-
+    BT_VLOG << "Created worker=" << pthread_self() << " bthread=" << g->main_tid()
+            << " tag=" << g->tag();
     tls_task_group = g;
     c->_nworkers << 1;
+    c->tag_nworkers(g->tag()) << 1;
     g->run_main_task();
 
     stat = g->main_stat();
@@ -87,10 +105,11 @@ void* TaskControl::worker_thread(void* arg) {
     tls_task_group = NULL;
     g->destroy_self();
     c->_nworkers << -1;
+    c->tag_nworkers(g->tag()) << -1;
     return NULL;
 }
 
-TaskGroup* TaskControl::create_group() {
+TaskGroup* TaskControl::create_group(bthread_tag_t tag) {
     TaskGroup* g = new (std::nothrow) TaskGroup(this);
     if (NULL == g) {
         LOG(FATAL) << "Fail to new TaskGroup";
@@ -101,7 +120,7 @@ TaskGroup* TaskControl::create_group() {
         delete g;
         return NULL;
     }
-    if (_add_group(g) != 0) {
+    if (_add_group(g, tag) != 0) {
         delete g;
         return NULL;
     }
@@ -117,6 +136,19 @@ static double get_cumulated_worker_time_from_this(void *arg) {
     return static_cast<TaskControl*>(arg)->get_cumulated_worker_time();
 }
 
+struct CumulatedWithTagArgs {
+    CumulatedWithTagArgs(TaskControl* _c, bthread_tag_t _t) : c(_c), t(_t) {}
+    TaskControl* c;
+    bthread_tag_t t;
+};
+
+static double get_cumulated_worker_time_from_this_with_tag(void* arg) {
+    auto a = static_cast<CumulatedWithTagArgs*>(arg);
+    auto c = a->c;
+    auto t = a->t;
+    return c->get_cumulated_worker_time_with_tag(t);
+}
+
 static int64_t get_cumulated_switch_count_from_this(void *arg) {
     return static_cast<TaskControl*>(arg)->get_cumulated_switch_count();
 }
@@ -127,8 +159,9 @@ static int64_t get_cumulated_signal_count_from_this(void *arg) {
 
 TaskControl::TaskControl()
     // NOTE: all fileds must be initialized before the vars.
-    : _ngroup(0)
-    , _groups((TaskGroup**)calloc(BTHREAD_MAX_CONCURRENCY, sizeof(TaskGroup*)))
+    : _tagged_ngroup(FLAGS_task_group_ntags)
+    , _tagged_groups(FLAGS_task_group_ntags)
+    , _init(false)
     , _stop(false)
     , _concurrency(0)
     , _next_worker_id(0)
@@ -144,10 +177,8 @@ TaskControl::TaskControl()
     , _signal_per_second(&_cumulated_signal_count)
     , _status(print_rq_sizes_in_the_tc, this)
     , _nbthreads("bthread_count")
-{
-    // calloc shall set memory to zero
-    CHECK(_groups) << "Fail to create array of groups";
-}
+    , _pl(FLAGS_task_group_ntags)
+{}
 
 int TaskControl::init(int concurrency) {
     if (_concurrency != 0) {
@@ -160,6 +191,18 @@ int TaskControl::init(int concurrency) {
     }
     _concurrency = concurrency;
 
+    // task group group by tags
+    for (int i = 0; i < FLAGS_task_group_ntags; ++i) {
+        _tagged_ngroup[i].store(0, std::memory_order_relaxed);
+        auto tag_str = std::to_string(i);
+        _tagged_nworkers.push_back(new bvar::Adder<int64_t>("bthread_worker_count", tag_str));
+        _tagged_cumulated_worker_time.push_back(new bvar::PassiveStatus<double>(
+            get_cumulated_worker_time_from_this_with_tag, new CumulatedWithTagArgs{this, i}));
+        _tagged_worker_usage_second.push_back(new bvar::PerSecond<bvar::PassiveStatus<double>>(
+            "bthread_worker_usage", tag_str, _tagged_cumulated_worker_time[i], 1));
+        _tagged_nbthreads.push_back(new bvar::Adder<int64_t>("bthread_count", tag_str));
+    }
+
     // Make sure TimerThread is ready.
     if (get_or_create_global_timer_thread() == NULL) {
         LOG(ERROR) << "Fail to get global_timer_thread";
@@ -168,8 +211,10 @@ int TaskControl::init(int concurrency) {
     
     _workers.resize(_concurrency);   
     for (int i = 0; i < _concurrency; ++i) {
-        const int rc = pthread_create(&_workers[i], NULL, worker_thread, this);
+        auto arg = new WorkerThreadArgs(this, i % FLAGS_task_group_ntags);
+        const int rc = pthread_create(&_workers[i], NULL, worker_thread, arg);
         if (rc) {
+            delete arg;
             LOG(ERROR) << "Fail to create _workers[" << i << "], " << berror(rc);
             return -1;
         }
@@ -182,13 +227,20 @@ int TaskControl::init(int concurrency) {
     // Wait for at least one group is added so that choose_one_group()
     // never returns NULL.
     // TODO: Handle the case that worker quits before add_group
-    while (_ngroup == 0) {
-        usleep(100);  // TODO: Elaborate
+    for (int i = 0; i < FLAGS_task_group_ntags;) {
+        if (_tagged_ngroup[i].load(std::memory_order_acquire) == 0) {
+            usleep(100);  // TODO: Elaborate
+            continue;
+        }
+        ++i;
     }
+
+    _init.store(true, butil::memory_order_release);
+
     return 0;
 }
 
-int TaskControl::add_workers(int num) {
+int TaskControl::add_workers(int num, bthread_tag_t tag) {
     if (num <= 0) {
         return 0;
     }
@@ -202,9 +254,11 @@ int TaskControl::add_workers(int num) {
         // Worker will add itself to _idle_workers, so we have to add
         // _concurrency before create a worker.
         _concurrency.fetch_add(1);
+        auto arg = new WorkerThreadArgs(this, tag);
         const int rc = pthread_create(
-                &_workers[i + old_concurency], NULL, worker_thread, this);
+                &_workers[i + old_concurency], NULL, worker_thread, arg);
         if (rc) {
+            delete arg;
             LOG(WARNING) << "Fail to create _workers[" << i + old_concurency
                          << "], " << berror(rc);
             _concurrency.fetch_sub(1, butil::memory_order_release);
@@ -216,10 +270,12 @@ int TaskControl::add_workers(int num) {
     return _concurrency.load(butil::memory_order_relaxed) - old_concurency;
 }
 
-TaskGroup* TaskControl::choose_one_group() {
-    const size_t ngroup = _ngroup.load(butil::memory_order_acquire);
+TaskGroup* TaskControl::choose_one_group(bthread_tag_t tag) {
+    CHECK(tag >= BTHREAD_TAG_DEFAULT && tag < FLAGS_task_group_ntags);
+    auto& groups = tag_group(tag);
+    const auto ngroup = tag_ngroup(tag).load(butil::memory_order_acquire);
     if (ngroup != 0) {
-        return _groups[butil::fast_rand_less_than(ngroup)];
+        return groups[butil::fast_rand_less_than(ngroup)];
     }
     CHECK(false) << "Impossible: ngroup is 0";
     return NULL;
@@ -236,10 +292,14 @@ void TaskControl::stop_and_join() {
     {
         BAIDU_SCOPED_LOCK(_modify_group_mutex);
         _stop = true;
-        _ngroup.exchange(0, butil::memory_order_relaxed); 
+        std::for_each(
+            _tagged_ngroup.begin(), _tagged_ngroup.end(),
+            [](butil::atomic<size_t>& index) { index.store(0, butil::memory_order_relaxed); });
     }
-    for (int i = 0; i < PARKING_LOT_NUM; ++i) {
-        _pl[i].stop();
+    for (int i = 0; i < FLAGS_task_group_ntags; ++i) {
+        for (auto& pl : _pl[i]) {
+            pl.stop();
+        }
     }
     // Interrupt blocking operations.
     for (size_t i = 0; i < _workers.size(); ++i) {
@@ -261,12 +321,9 @@ TaskControl::~TaskControl() {
     _status.hide();
     
     stop_and_join();
-
-    free(_groups);
-    _groups = NULL;
 }
 
-int TaskControl::_add_group(TaskGroup* g) {
+int TaskControl::_add_group(TaskGroup* g, bthread_tag_t tag) {
     if (__builtin_expect(NULL == g, 0)) {
         return -1;
     }
@@ -274,15 +331,17 @@ int TaskControl::_add_group(TaskGroup* g) {
     if (_stop) {
         return -1;
     }
-    size_t ngroup = _ngroup.load(butil::memory_order_relaxed);
+    g->set_tag(tag);
+    g->set_pl(&_pl[tag][butil::fmix64(pthread_numeric_id()) % PARKING_LOT_NUM]);
+    size_t ngroup = _tagged_ngroup[tag].load(butil::memory_order_relaxed);
     if (ngroup < (size_t)BTHREAD_MAX_CONCURRENCY) {
-        _groups[ngroup] = g;
-        _ngroup.store(ngroup + 1, butil::memory_order_release);
+        _tagged_groups[tag][ngroup] = g;
+        _tagged_ngroup[tag].store(ngroup + 1, butil::memory_order_release);
     }
     mu.unlock();
     // See the comments in _destroy_group
     // TODO: Not needed anymore since non-worker pthread cannot have TaskGroup
-    signal_task(65536);
+    // signal_task(65536, tag);
     return 0;
 }
 
@@ -303,11 +362,13 @@ int TaskControl::_destroy_group(TaskGroup* g) {
     bool erased = false;
     {
         BAIDU_SCOPED_LOCK(_modify_group_mutex);
-        const size_t ngroup = _ngroup.load(butil::memory_order_relaxed);
+        auto tag = g->tag();
+        auto& groups = tag_group(tag);
+        const size_t ngroup = tag_ngroup(tag).load(butil::memory_order_relaxed);
         for (size_t i = 0; i < ngroup; ++i) {
-            if (_groups[i] == g) {
+            if (groups[i] == g) {
                 // No need for atomic_thread_fence because lock did it.
-                _groups[i] = _groups[ngroup - 1];
+                groups[i] = groups[ngroup - 1];
                 // Change _ngroup and keep _groups unchanged at last so that:
                 //  - If steal_task sees the newest _ngroup, it would not touch
                 //    _groups[ngroup -1]
@@ -317,7 +378,7 @@ int TaskControl::_destroy_group(TaskGroup* g) {
                 //    overwrite it, since we do signal_task in _add_group(),
                 //    we think the pending tasks of _groups[ngroup - 1] would
                 //    not miss.
-                _ngroup.store(ngroup - 1, butil::memory_order_release);
+                tag_ngroup(tag).store(ngroup - 1, butil::memory_order_release);
                 //_groups[ngroup - 1] = NULL;
                 erased = true;
                 break;
@@ -339,9 +400,10 @@ int TaskControl::_destroy_group(TaskGroup* g) {
 }
 
 bool TaskControl::steal_task(bthread_t* tid, size_t* seed, size_t offset) {
+    auto tag = tls_task_group->tag();
     // 1: Acquiring fence is paired with releasing fence in _add_group to
     // avoid accessing uninitialized slot of _groups.
-    const size_t ngroup = _ngroup.load(butil::memory_order_acquire/*1*/);
+    const size_t ngroup = tag_ngroup(tag).load(butil::memory_order_acquire/*1*/);
     if (0 == ngroup) {
         return false;
     }
@@ -349,8 +411,9 @@ bool TaskControl::steal_task(bthread_t* tid, size_t* seed, size_t offset) {
     // NOTE: Don't return inside `for' iteration since we need to update |seed|
     bool stolen = false;
     size_t s = *seed;
+    auto& groups = tag_group(tag);
     for (size_t i = 0; i < ngroup; ++i, s += offset) {
-        TaskGroup* g = _groups[s % ngroup];
+        TaskGroup* g = groups[s % ngroup];
         // g is possibly NULL because of concurrent _destroy_group
         if (g) {
             if (g->_rq.steal(tid)) {
@@ -367,7 +430,7 @@ bool TaskControl::steal_task(bthread_t* tid, size_t* seed, size_t offset) {
     return stolen;
 }
 
-void TaskControl::signal_task(int num_task) {
+void TaskControl::signal_task(int num_task, bthread_tag_t tag) {
     if (num_task <= 0) {
         return;
     }
@@ -378,14 +441,15 @@ void TaskControl::signal_task(int num_task) {
     if (num_task > 2) {
         num_task = 2;
     }
+    auto& pl = tag_pl(tag);
     int start_index = butil::fmix64(pthread_numeric_id()) % PARKING_LOT_NUM;
-    num_task -= _pl[start_index].signal(1);
+    num_task -= pl[start_index].signal(1);
     if (num_task > 0) {
         for (int i = 1; i < PARKING_LOT_NUM && num_task > 0; ++i) {
             if (++start_index >= PARKING_LOT_NUM) {
                 start_index = 0;
             }
-            num_task -= _pl[start_index].signal(1);
+            num_task -= pl[start_index].signal(1);
         }
     }
     if (num_task > 0 &&
@@ -394,21 +458,26 @@ void TaskControl::signal_task(int num_task) {
         // TODO: Reduce this lock
         BAIDU_SCOPED_LOCK(g_task_control_mutex);
         if (_concurrency.load(butil::memory_order_acquire) < FLAGS_bthread_concurrency) {
-            add_workers(1);
+            add_workers(1, tag);
         }
     }
 }
 
 void TaskControl::print_rq_sizes(std::ostream& os) {
-    const size_t ngroup = _ngroup.load(butil::memory_order_relaxed);
+    size_t ngroup = 0;
+    std::for_each(_tagged_ngroup.begin(), _tagged_ngroup.end(), [&](butil::atomic<size_t>& index) {
+        ngroup += index.load(butil::memory_order_relaxed);
+    });
     DEFINE_SMALL_ARRAY(int, nums, ngroup, 128);
     {
         BAIDU_SCOPED_LOCK(_modify_group_mutex);
         // ngroup > _ngroup: nums[_ngroup ... ngroup-1] = 0
         // ngroup < _ngroup: just ignore _groups[_ngroup ... ngroup-1]
-        for (size_t i = 0; i < ngroup; ++i) {
-            nums[i] = (_groups[i] ? _groups[i]->_rq.volatile_size() : 0);
-        }
+        int i = 0;
+        for_each_task_group([&](TaskGroup* g) {
+            nums[i] = (g ? g->_rq.volatile_size() : 0);
+            ++i;
+        });
     }
     for (size_t i = 0; i < ngroup; ++i) {
         os << nums[i] << ' ';
@@ -418,10 +487,22 @@ void TaskControl::print_rq_sizes(std::ostream& os) {
 double TaskControl::get_cumulated_worker_time() {
     int64_t cputime_ns = 0;
     BAIDU_SCOPED_LOCK(_modify_group_mutex);
-    const size_t ngroup = _ngroup.load(butil::memory_order_relaxed);
+    for_each_task_group([&](TaskGroup* g) {
+        if (g) {
+            cputime_ns += g->_cumulated_cputime_ns;
+        }
+    });
+    return cputime_ns / 1000000000.0;
+}
+
+double TaskControl::get_cumulated_worker_time_with_tag(bthread_tag_t tag) {
+    int64_t cputime_ns = 0;
+    BAIDU_SCOPED_LOCK(_modify_group_mutex);
+    const size_t ngroup = tag_ngroup(tag).load(butil::memory_order_relaxed);
+    auto& groups = tag_group(tag);
     for (size_t i = 0; i < ngroup; ++i) {
-        if (_groups[i]) {
-            cputime_ns += _groups[i]->_cumulated_cputime_ns;
+        if (groups[i]) {
+            cputime_ns += groups[i]->_cumulated_cputime_ns;
         }
     }
     return cputime_ns / 1000000000.0;
@@ -430,25 +511,22 @@ double TaskControl::get_cumulated_worker_time() {
 int64_t TaskControl::get_cumulated_switch_count() {
     int64_t c = 0;
     BAIDU_SCOPED_LOCK(_modify_group_mutex);
-    const size_t ngroup = _ngroup.load(butil::memory_order_relaxed);
-    for (size_t i = 0; i < ngroup; ++i) {
-        if (_groups[i]) {
-            c += _groups[i]->_nswitch;
+    for_each_task_group([&](TaskGroup* g) {
+        if (g) {
+            c += g->_nswitch;
         }
-    }
+    });
     return c;
 }
 
 int64_t TaskControl::get_cumulated_signal_count() {
     int64_t c = 0;
     BAIDU_SCOPED_LOCK(_modify_group_mutex);
-    const size_t ngroup = _ngroup.load(butil::memory_order_relaxed);
-    for (size_t i = 0; i < ngroup; ++i) {
-        TaskGroup* g = _groups[i];
+    for_each_task_group([&](TaskGroup* g) {
         if (g) {
             c += g->_nsignaled + g->_remote_nsignaled;
         }
-    }
+    });
     return c;
 }
 
