@@ -4,7 +4,9 @@
 
 #include "butil/basictypes.h"
 #include "butil/logging.h"
-
+#include "butil/gperftools_profiler.h"
+#include "butil/files/temp_file.h"
+#include "butil/popen.h"
 #include <gtest/gtest.h>
 #include <gflags/gflags.h>
 
@@ -14,6 +16,7 @@ namespace logging {
 DECLARE_bool(crash_on_fatal_log);
 DECLARE_int32(v);
 DECLARE_bool(log_func_name);
+DECLARE_bool(async_log);
 
 namespace {
 
@@ -475,6 +478,163 @@ TEST_F(LoggingTest, log_func) {
     ::logging::FLAGS_crash_on_fatal_log = old_crash_on_fatal_log;
 }
 
+bool g_started = false;
+bool g_stopped = false;
+int g_prof_name_counter = 0;
+butil::atomic<uint64_t> test_logging_count(0);
+
+void* test_async_log(void* arg) {
+    if (arg == NULL) {
+        return NULL;
+    }
+    auto log = (std::string*)(arg);
+    while (!g_stopped) {
+        LOG(INFO) << *log;
+        test_logging_count.fetch_add(1);
+    }
+
+    return NULL;
+}
+
+TEST_F(LoggingTest, async_log) {
+    bool saved_async_log = FLAGS_async_log;
+    FLAGS_async_log = true;
+    butil::TempFile temp_file;
+    LoggingSettings settings;
+    settings.logging_dest = LOG_TO_FILE;
+    settings.log_file = temp_file.fname();
+    settings.delete_old = DELETE_OLD_LOG_FILE;
+    InitLogging(settings);
+
+    std::string log = "135792468";
+    int thread_num = 8;
+    pthread_t threads[thread_num];
+    for (int i = 0; i < thread_num; ++i) {
+        ASSERT_EQ(0, pthread_create(&threads[i], NULL, test_async_log, &log));
+    }
+
+    sleep(5);
+
+    g_stopped = true;
+    for (int i = 0; i < thread_num; ++i) {
+        pthread_join(threads[i], NULL);
+    }
+    // Wait for async log thread to flush all logs to file.
+    sleep(10);
+
+    std::ostringstream oss;
+    std::string cmd = butil::string_printf("grep -c %s %s",
+        log.c_str(), temp_file.fname());
+    ASSERT_LE(0, butil::read_command_output(oss, cmd.c_str()));
+    uint64_t log_count = std::strtol(oss.str().c_str(), NULL, 10);
+    ASSERT_EQ(log_count, test_logging_count.load());
+
+    FLAGS_async_log = saved_async_log;
+}
+
+struct BAIDU_CACHELINE_ALIGNMENT PerfArgs {
+    const std::string* log;
+    int64_t counter;
+    int64_t elapse_ns;
+    bool ready;
+
+    PerfArgs() : log(NULL), counter(0), elapse_ns(0), ready(false) {}
+};
+
+void* test_log(void* void_arg) {
+    auto args = (PerfArgs*)void_arg;
+    args->ready = true;
+    butil::Timer t;
+    while (!g_stopped) {
+        if (g_started) {
+            break;
+        }
+        usleep(10);
+    }
+    t.start();
+    while (!g_stopped) {
+        {
+            LOG(INFO) << *args->log;
+            test_logging_count.fetch_add(1, butil::memory_order_relaxed);
+        }
+        ++args->counter;
+    }
+    t.stop();
+    args->elapse_ns = t.n_elapsed();
+    return NULL;
+}
+
+void PerfTest(int thread_num, const std::string& log, bool async) {
+    FLAGS_async_log = async;
+
+    g_started = false;
+    g_stopped = false;
+    pthread_t threads[thread_num];
+    std::vector<PerfArgs> args(thread_num);
+    for (int i = 0; i < thread_num; ++i) {
+        args[i].log = &log;
+        ASSERT_EQ(0, pthread_create(&threads[i], NULL, test_log, &args[i]));
+    }
+    while (true) {
+        bool all_ready = true;
+        for (int i = 0; i < thread_num; ++i) {
+            if (!args[i].ready) {
+                all_ready = false;
+                break;
+            }
+        }
+        if (all_ready) {
+            break;
+        }
+        usleep(1000);
+    }
+    g_started = true;
+    char prof_name[32];
+    snprintf(prof_name, sizeof(prof_name), "logging_%d.prof", ++g_prof_name_counter);
+    ProfilerStart(prof_name);
+    sleep(5);
+    ProfilerStop();
+    g_stopped = true;
+    int64_t wait_time = 0;
+    int64_t count = 0;
+    for (int i = 0; i < thread_num; ++i) {
+        pthread_join(threads[i], NULL);
+        wait_time += args[i].elapse_ns;
+        count += args[i].counter;
+    }
+    std::cout << " thread_num=" << thread_num
+              << " log_type=" << (async ? "async" : "sync")
+              << " log_size=" << log.size()
+              << " count=" << count
+              << " average_time=" << wait_time / (double)count
+              << std::endl;
+}
+
+TEST_F(LoggingTest, performance) {
+    bool saved_async_log = FLAGS_async_log;
+
+    LoggingSettings settings;
+    settings.logging_dest = LOG_TO_FILE;
+    InitLogging(settings);
+    std::string log(64, 'a');
+    int thread_num = 1;
+    PerfTest(thread_num, log, true);
+    sleep(10);
+    PerfTest(thread_num, log, false);
+
+    thread_num = 2;
+    PerfTest(thread_num, log, true);
+    sleep(10);
+    PerfTest(thread_num, log, false);
+
+    thread_num = 4;
+    PerfTest(thread_num, log, true);
+    sleep(10);
+    PerfTest(thread_num, log, false);
+
+    FLAGS_async_log = saved_async_log;
+}
+
 }  // namespace
 
 }  // namespace logging
